@@ -1,5 +1,6 @@
 from __future__ import absolute_import
 import time
+import hashlib
 
 import numpy as nm
 import warnings
@@ -28,6 +29,39 @@ def solve(mtx, rhs, solver_class=None, solver_conf=None):
 
     return solution
 
+def _get_cs_matrix_hash(mtx, chunk_size=100000):
+    def _gen_array_chunks(arr):
+        ii = 0
+        while len(arr[ii:]):
+            yield arr[ii:ii+chunk_size].tobytes()
+            ii += chunk_size
+
+    sha1 = hashlib.sha1()
+    for chunk in _gen_array_chunks(mtx.indptr):
+        sha1.update(chunk)
+    for chunk in _gen_array_chunks(mtx.indices):
+        sha1.update(chunk)
+    for chunk in _gen_array_chunks(mtx.data):
+        sha1.update(chunk)
+
+    digest = sha1.hexdigest()
+    return digest
+
+def _is_new_matrix(mtx, mtx_digest, force_reuse=False):
+    if not isinstance(mtx, sps.csr_matrix):
+        return True, mtx_digest
+
+    if force_reuse:
+        return False, mtx_digest
+
+    id0, digest0 = mtx_digest
+    id1 = id(mtx)
+    digest1 = _get_cs_matrix_hash(mtx)
+    if (id1 == id0) and (digest1 == digest0):
+        return False, (id1, digest1)
+
+    return True, (id1, digest1)
+
 def standard_call(call):
     """
     Decorator handling argument preparation and timing for linear solvers.
@@ -117,7 +151,7 @@ class ScipyDirect(LinearSolver):
     ]
 
     def __init__(self, conf, **kwargs):
-        LinearSolver.__init__(self, conf, mtx_id=None, solve=None, **kwargs)
+        LinearSolver.__init__(self, conf, solve=None, **kwargs)
         um = self.sls = None
 
         aux = try_imports(['import scipy.linsolve as sls',
@@ -164,9 +198,10 @@ class ScipyDirect(LinearSolver):
             return self.sls.spsolve(mtx, rhs)
 
     def presolve(self, mtx):
-        if self.mtx_id != id(mtx):
+        is_new, mtx_digest = _is_new_matrix(mtx, self.mtx_digest)
+        if is_new:
             self.solve = self.sls.factorized(mtx)
-            self.mtx_id = id(mtx)
+            self.mtx_digest = mtx_digest
 
 class ScipyIterative(LinearSolver):
     """
@@ -195,7 +230,7 @@ class ScipyIterative(LinearSolver):
          """),
         ('i_max', 'int', 100, False,
          'The maximum number of iterations.'),
-        ('eps_a', 'float', 1e0, False,
+        ('eps_a', 'float', 1e-8, False,
          'The absolute tolerance for the residual.'),
         ('eps_r', 'float', 1e-8, False,
          'The relative tolerance for the residual.'),
@@ -230,6 +265,7 @@ class ScipyIterative(LinearSolver):
                  i_max=None, mtx=None, status=None, context=None, **kwargs):
         solver_kwargs = self.build_solver_kwargs(conf)
 
+        eps_a = get_default(eps_a, self.conf.eps_a)
         eps_r = get_default(eps_r, self.conf.eps_r)
         i_max = get_default(i_max, self.conf.i_max)
 
@@ -267,7 +303,7 @@ class ScipyIterative(LinearSolver):
         solver_kwargs.update(prec_args)
 
         try:
-            sol, info = self.solver(mtx, rhs, x0=x0, atol=eps_a, rtol=eps_r,
+            sol, info = self.solver(mtx, rhs, x0=x0, atol=eps_a, tol=eps_r,
                                     maxiter=i_max, callback=iter_callback,
                                     **solver_kwargs)
         except TypeError:
@@ -308,6 +344,9 @@ class PyAMGSolver(LinearSolver):
          'The maximum number of iterations.'),
         ('eps_r', 'float', 1e-8, False,
          'The relative tolerance for the residual.'),
+        ('force_reuse', 'bool', False, False,
+         """If True, skip the check whether the MG solver object corresponds
+            to the `mtx` argument: it is always reused."""),
         ('*', '*', None, False,
          """Additional parameters supported by the method. Use the 'method:'
             prefix for arguments of the method construction function
@@ -326,7 +365,7 @@ class PyAMGSolver(LinearSolver):
             msg =  'cannot import pyamg!'
             raise ImportError(msg)
 
-        LinearSolver.__init__(self, conf, mtx_id=None, mg=None, **kwargs)
+        LinearSolver.__init__(self, conf, mg=None, **kwargs)
 
         try:
             solver = getattr(pyamg, self.conf.method)
@@ -365,12 +404,14 @@ class PyAMGSolver(LinearSolver):
             # Call an optional user-defined callback.
             callback(sol)
 
-        if self.mtx_id != id(mtx):
+        is_new, mtx_digest = _is_new_matrix(mtx, self.mtx_digest,
+                                            force_reuse=conf.force_reuse)
+        if is_new or (self.mg is None):
             _kwargs = {key[7:] : val
                        for key, val in six.iteritems(solver_kwargs)
                        if key.startswith('method:')}
             self.mg = self.solver(mtx, **_kwargs)
-            self.mtx_id = id(mtx)
+            self.mtx_digest = mtx_digest
 
         _kwargs = {key[6:] : val
                    for key, val in six.iteritems(solver_kwargs)
@@ -422,7 +463,7 @@ class PyAMGKrylovSolver(LinearSolver):
             msg =  'cannot import pyamg.krylov!'
             raise ImportError(msg)
 
-        LinearSolver.__init__(self, conf, mtx_id=None, mg=None,
+        LinearSolver.__init__(self, conf, mg=None,
                               context=context, **kwargs)
 
         try:
@@ -529,6 +570,12 @@ class PETScKrylovSolver(LinearSolver):
          'The relative tolerance for the residual.'),
         ('eps_d', 'float', 1e5, False,
          'The divergence tolerance for the residual.'),
+        ('force_reuse', 'bool', False, False,
+         """If True, skip the check whether the KSP solver object corresponds
+            to the `mtx` argument: it is always reused."""),
+        ('*', '*', None, False,
+         """Additional parameters supported by the method. Can be used to pass
+            all PETSc options supported by :func:`petsc.Options()`."""),
     ]
 
     _precond_sides = {None : None, 'left' : 0, 'right' : 1, 'symmetric' : 2}
@@ -546,7 +593,7 @@ class PETScKrylovSolver(LinearSolver):
 
         LinearSolver.__init__(self, conf, petsc=petsc, comm=comm,
                               converged_reasons=converged_reasons,
-                              fields=None, mtx_id=0, ksp=None, pmtx=None,
+                              fields=None, ksp=None, pmtx=None,
                               context=context, **kwargs)
 
     def set_field_split(self, field_ranges, comm=None):
@@ -568,10 +615,13 @@ class PETScKrylovSolver(LinearSolver):
                                                     comm=comm)
             self.fields.append((key, field_is))
 
-    def create_ksp(self, comm=None):
+    def create_ksp(self, options=None, comm=None):
         optDB = self.petsc.Options()
 
         optDB['sub_pc_type'] = self.conf.sub_precond
+        if options is not None:
+            for key, val in six.iteritems(options):
+                optDB[key] = val
 
         ksp = self.petsc.KSP()
         ksp.create(comm)
@@ -616,19 +666,23 @@ class PETScKrylovSolver(LinearSolver):
     def __call__(self, rhs, x0=None, conf=None, eps_a=None, eps_r=None,
                  i_max=None, mtx=None, status=None, comm=None, context=None,
                  **kwargs):
+        solver_kwargs = self.build_solver_kwargs(conf)
+
         eps_a = get_default(eps_a, self.conf.eps_a)
         eps_r = get_default(eps_r, self.conf.eps_r)
         i_max = get_default(i_max, self.conf.i_max)
         eps_d = self.conf.eps_d
 
-        if self.mtx_id == id(mtx):
+        is_new, mtx_digest = _is_new_matrix(mtx, self.mtx_digest,
+                                            force_reuse=conf.force_reuse)
+        if (not is_new) and self.ksp is not None:
             ksp = self.ksp
             pmtx = self.pmtx
 
         else:
             pmtx = self.create_petsc_matrix(mtx, comm=comm)
 
-            ksp = self.create_ksp(comm=comm)
+            ksp = self.create_ksp(options=solver_kwargs, comm=comm)
             ksp.setOperators(pmtx)
             ksp.setTolerances(atol=eps_a, rtol=eps_r, divtol=eps_d,
                               max_it=i_max)
@@ -638,7 +692,7 @@ class PETScKrylovSolver(LinearSolver):
                 ksp.pc.setPythonContext(setup_precond(mtx, context))
 
             ksp.setFromOptions()
-            self.mtx_id = id(mtx)
+            self.mtx_digest = mtx_digest
             self.ksp = ksp
             self.pmtx = pmtx
 
@@ -679,6 +733,127 @@ class PETScKrylovSolver(LinearSolver):
 
         return sol
 
+
+class MUMPSSolver(LinearSolver):
+    """
+    Interface to MUMPS solver.
+
+    """
+    name = 'ls.mumps'
+
+    __metaclass__ = SolverMeta
+
+    _parameters = []
+
+    def __init__(self, conf, **kwargs):
+        import sfepy.solvers.ls_mumps as mumps
+
+        self.mumps_ls = None
+        mumps.load_mumps_libraries()  # try to load MUMPS libraries
+
+        LinearSolver.__init__(self, conf, mumps=mumps, mumps_ls=None,
+                              mumps_presolved=False, **kwargs)
+
+    @standard_call
+    def __call__(self, rhs, x0=None, conf=None, eps_a=None, eps_r=None,
+                 i_max=None, mtx=None, status=None, **kwargs):
+
+        if self.mumps_ls is None:
+            system = 'complex' if mtx.dtype.name.startswith('complex')\
+                else 'real'
+            self.mumps_ls = self.mumps.MumpsSolver(system=system)
+
+        if not self.mumps_presolved:
+            self.presolve(mtx)
+
+        out = rhs.copy()
+        self.mumps_ls.set_b(out)
+        self.mumps_ls(3)  # solve
+        from scipy.io import savemat
+        savemat('test_matrix_vector.mat', {'A': mtx, 'b': rhs, 'x': out})
+
+        return out
+
+    def presolve(self, mtx):
+        is_new, mtx_digest = _is_new_matrix(mtx, self.mtx_digest)
+        if is_new:
+            if self.conf.verbose:
+                self.mumps_ls.set_verbose()
+
+            self.mumps_ls.set_A_centralized(mtx)
+            self.mumps_ls(1)  # analyze
+            self.mumps_ls(2)  # factorize
+            self.mumps_presolved = True
+            self.mtx_digest = mtx_digest
+
+    def __del__(self):
+        if self.mumps_ls is not None:
+            del(self.mumps_ls)
+
+
+class MUMPSParallelSolver(LinearSolver):
+    """
+    Interface to MUMPS parallel solver.
+    """
+    name = 'ls.mumps_par'
+
+    __metaclass__ = SolverMeta
+
+    _parameters = []
+
+    def __init__(self, conf, **kwargs):
+        import multiprocessing
+        import sfepy.solvers.ls_mumps as mumps
+
+        mumps.load_mumps_libraries()  # try to load MUMPS libraries
+
+        LinearSolver.__init__(self, conf, mumps=mumps, mumps_ls=None,
+                              number_of_cpu=multiprocessing.cpu_count(),
+                              mumps_presolved=False, **kwargs)
+
+    @standard_call
+    def __call__(self, rhs, x0=None, conf=None, eps_a=None, eps_r=None,
+                 i_max=None, mtx=None, status=None, **kwargs):
+        from mpi4py import MPI
+        import sys
+        from sfepy import data_dir
+        import os.path as op
+        from tempfile import gettempdir
+
+        def tmpfile(fname):
+            return op.join(gettempdir(), fname)
+
+        mtx_coo = mtx.tocoo()
+        n = mtx.shape[0]
+        nz = mtx_coo.row.shape[0]
+        flags = nm.memmap(tmpfile('vals_falgs.array'), dtype='int32',
+                          mode='w+', shape=(2,))
+        flags[0] = n
+        flags[1] = 1 if mtx_coo.data.dtype.name.startswith('complex') else 0
+        idxs = nm.memmap(tmpfile('idxs.array'), dtype='int32',
+                         mode='w+', shape=(2, nz))
+        idxs[0, :] = mtx_coo.row + 1
+        idxs[1, :] = mtx_coo.col + 1
+
+        dtype = {0: 'float64', 1: 'complex128'}[flags[1]]
+        vals_mtx = nm.memmap(tmpfile('vals_mtx.array'), dtype=dtype,
+                             mode='w+', shape=(nz,))
+        vals_rhs = nm.memmap(tmpfile('vals_rhs.array'), dtype=dtype,
+                             mode='w+', shape=(n,))
+        vals_mtx[:] = mtx_coo.data
+        vals_rhs[:] = rhs
+
+        mumps_call = op.join(data_dir, 'sfepy', 'solvers',
+                             'ls_mumps_parallel.py')
+        comm = MPI.COMM_SELF.Spawn(sys.executable, args=[mumps_call],
+                                   maxprocs=self.number_of_cpu)
+        comm.Disconnect()
+
+        out = nm.memmap(tmpfile('vals_x.array'), dtype=dtype, mode='r')
+
+        return out
+
+
 class SchurGeneralized(ScipyDirect):
     r"""
     Generalized Schur complement.
@@ -698,14 +873,17 @@ class SchurGeneralized(ScipyDirect):
     ]
 
     def __init__(self, conf, context=None, **kwargs):
-        from sfepy.discrete.state import State
-
         ScipyDirect.__init__(self, conf, context=context, **kwargs)
 
-        equations = context.equations
+    @standard_call
+    def __call__(self, rhs, x0=None, conf=None, eps_a=None, eps_r=None,
+                 i_max=None, mtx=None, status=None, **kwargs):
+        from sfepy.discrete.state import State
+
+        equations = self.context.equations
         aux_state = State(equations.variables)
 
-        conf.idxs = {}
+        mtxi = {}
         for bk, bv in six.iteritems(conf.blocks):
             aux_state.fill(0.0)
             for jj in bv:
@@ -714,13 +892,8 @@ class SchurGeneralized(ScipyDirect):
 
             aux_state.apply_ebc()
             vec0 = aux_state.get_reduced()
-            conf.idxs[bk] = nm.where(nm.isnan(vec0))[0]
+            mtxi[bk] = nm.where(nm.isnan(vec0))[0]
 
-    @standard_call
-    def __call__(self, rhs, x0=None, conf=None, eps_a=None, eps_r=None,
-                 i_max=None, mtx=None, status=None, **kwargs):
-
-        mtxi= self.orig_conf.idxs
         mtxslc_s = {}
         mtxslc_f = {}
         nn = {}
@@ -799,7 +972,7 @@ class SchurComplement(SchurGeneralized):
 
     __metaclass__ = SolverMeta
 
-    _parameters = ScipyDirect._parameters + [
+    _parameters = SchurGeneralized._parameters + [
         ('eliminate', 'list', None, True,
          'The list of variables to eliminate.'),
         ('keep', 'list', None, True,
@@ -851,13 +1024,14 @@ class MultiProblem(ScipyDirect):
     ]
 
     def __init__(self, conf, context=None, **kwargs):
+        ScipyDirect.__init__(self, conf, context=context, **kwargs)
+
+    def init_subproblems(self, conf, **kwargs):
         from sfepy.discrete.state import State
         from sfepy.discrete import Problem
         from sfepy.base.conf import ProblemConf, get_standard_keywords
         from scipy.spatial import cKDTree as KDTree
 
-        ScipyDirect.__init__(self, conf, context=context, **kwargs)
-
         # init subproblems
         problem = self.context
         pb_vars = problem.get_variables()
@@ -989,6 +1163,7 @@ class MultiProblem(ScipyDirect):
     @standard_call
     def __call__(self, rhs, x0=None, conf=None, eps_a=None, eps_r=None,
                  i_max=None, mtx=None, status=None, **kwargs):
+        self.init_subproblems(self.conf, **kwargs)
 
         max_indx = 0
         hst = nm.hstack
